
#include <stdio.h>
#include <string.h>   //strlen
#include <stdlib.h>   //strlen
#include <stdint.h>
#include <sys/socket.h>
#include <arpa/inet.h>   //inet_addr
#include <unistd.h>   //write
//#include <netinet/in.h> not sure this is needed
#include <pthread.h> //for threading , link with lpthread
#include <time.h>

//$ gcc program.c -lpthread -o NetDanceServer


typedef struct
{
    uint32_t state[5];
    uint32_t count[2];
    unsigned char buffer[64];
} SHA1_CTX;

void SHA1Transform(
    uint32_t state[5],
    const unsigned char buffer[64]
    );

void SHA1Init(
    SHA1_CTX * context
    );

void SHA1Update(
    SHA1_CTX * context,
    const unsigned char *data,
    uint32_t len
    );

void SHA1Final(
    unsigned char digest[20],
    SHA1_CTX * context
    );

void SHA1(
    char *hash_out,
    const char *str,
    int len);

/* #define LITTLE_ENDIAN * This should be #define'd already, if true. */
/* #define SHA1HANDSOFF * Copies data before messing with it. */

#define SHA1HANDSOFF

/* for uint32_t */

#define rol(value, bits) (((value) << (bits)) | ((value) >> (32 - (bits))))

/* blk0() and blk() perform the initial expand. */
/* I got the idea of expanding during the round function from SSLeay */
#if BYTE_ORDER == LITTLE_ENDIAN
#define blk0(i) (block->l[i] = (rol(block->l[i],24)&0xFF00FF00) \
    |(rol(block->l[i],8)&0x00FF00FF))
#elif BYTE_ORDER == BIG_ENDIAN
#define blk0(i) block->l[i]
#else
#error "Endianness not defined!"
#endif
#define blk(i) (block->l[i&15] = rol(block->l[(i+13)&15]^block->l[(i+8)&15] \
    ^block->l[(i+2)&15]^block->l[i&15],1))

/* (R0+R1), R2, R3, R4 are the different operations used in SHA1 */
#define R0(v,w,x,y,z,i) z+=((w&(x^y))^y)+blk0(i)+0x5A827999+rol(v,5);w=rol(w,30);
#define R1(v,w,x,y,z,i) z+=((w&(x^y))^y)+blk(i)+0x5A827999+rol(v,5);w=rol(w,30);
#define R2(v,w,x,y,z,i) z+=(w^x^y)+blk(i)+0x6ED9EBA1+rol(v,5);w=rol(w,30);
#define R3(v,w,x,y,z,i) z+=(((w|x)&y)|(w&x))+blk(i)+0x8F1BBCDC+rol(v,5);w=rol(w,30);
#define R4(v,w,x,y,z,i) z+=(w^x^y)+blk(i)+0xCA62C1D6+rol(v,5);w=rol(w,30);


/* Hash a single 512-bit block. This is the core of the algorithm. */

void SHA1Transform(
    uint32_t state[5],
    const unsigned char buffer[64]
)
{
    uint32_t a, b, c, d, e;

    typedef union
    {
        unsigned char c[64];
        uint32_t l[16];
    } CHAR64LONG16;

#ifdef SHA1HANDSOFF
    CHAR64LONG16 block[1];      /* use array to appear as a pointer */

    memcpy(block, buffer, 64);
#else
    /* The following had better never be used because it causes the
     * pointer-to-const buffer to be cast into a pointer to non-const.
     * And the result is written through.  I threw a "const" in, hoping
     * this will cause a diagnostic.
     */
    CHAR64LONG16 *block = (const CHAR64LONG16 *) buffer;
#endif
    /* Copy context->state[] to working vars */
    a = state[0];
    b = state[1];
    c = state[2];
    d = state[3];
    e = state[4];
    /* 4 rounds of 20 operations each. Loop unrolled. */
    R0(a, b, c, d, e, 0);
    R0(e, a, b, c, d, 1);
    R0(d, e, a, b, c, 2);
    R0(c, d, e, a, b, 3);
    R0(b, c, d, e, a, 4);
    R0(a, b, c, d, e, 5);
    R0(e, a, b, c, d, 6);
    R0(d, e, a, b, c, 7);
    R0(c, d, e, a, b, 8);
    R0(b, c, d, e, a, 9);
    R0(a, b, c, d, e, 10);
    R0(e, a, b, c, d, 11);
    R0(d, e, a, b, c, 12);
    R0(c, d, e, a, b, 13);
    R0(b, c, d, e, a, 14);
    R0(a, b, c, d, e, 15);
    R1(e, a, b, c, d, 16);
    R1(d, e, a, b, c, 17);
    R1(c, d, e, a, b, 18);
    R1(b, c, d, e, a, 19);
    R2(a, b, c, d, e, 20);
    R2(e, a, b, c, d, 21);
    R2(d, e, a, b, c, 22);
    R2(c, d, e, a, b, 23);
    R2(b, c, d, e, a, 24);
    R2(a, b, c, d, e, 25);
    R2(e, a, b, c, d, 26);
    R2(d, e, a, b, c, 27);
    R2(c, d, e, a, b, 28);
    R2(b, c, d, e, a, 29);
    R2(a, b, c, d, e, 30);
    R2(e, a, b, c, d, 31);
    R2(d, e, a, b, c, 32);
    R2(c, d, e, a, b, 33);
    R2(b, c, d, e, a, 34);
    R2(a, b, c, d, e, 35);
    R2(e, a, b, c, d, 36);
    R2(d, e, a, b, c, 37);
    R2(c, d, e, a, b, 38);
    R2(b, c, d, e, a, 39);
    R3(a, b, c, d, e, 40);
    R3(e, a, b, c, d, 41);
    R3(d, e, a, b, c, 42);
    R3(c, d, e, a, b, 43);
    R3(b, c, d, e, a, 44);
    R3(a, b, c, d, e, 45);
    R3(e, a, b, c, d, 46);
    R3(d, e, a, b, c, 47);
    R3(c, d, e, a, b, 48);
    R3(b, c, d, e, a, 49);
    R3(a, b, c, d, e, 50);
    R3(e, a, b, c, d, 51);
    R3(d, e, a, b, c, 52);
    R3(c, d, e, a, b, 53);
    R3(b, c, d, e, a, 54);
    R3(a, b, c, d, e, 55);
    R3(e, a, b, c, d, 56);
    R3(d, e, a, b, c, 57);
    R3(c, d, e, a, b, 58);
    R3(b, c, d, e, a, 59);
    R4(a, b, c, d, e, 60);
    R4(e, a, b, c, d, 61);
    R4(d, e, a, b, c, 62);
    R4(c, d, e, a, b, 63);
    R4(b, c, d, e, a, 64);
    R4(a, b, c, d, e, 65);
    R4(e, a, b, c, d, 66);
    R4(d, e, a, b, c, 67);
    R4(c, d, e, a, b, 68);
    R4(b, c, d, e, a, 69);
    R4(a, b, c, d, e, 70);
    R4(e, a, b, c, d, 71);
    R4(d, e, a, b, c, 72);
    R4(c, d, e, a, b, 73);
    R4(b, c, d, e, a, 74);
    R4(a, b, c, d, e, 75);
    R4(e, a, b, c, d, 76);
    R4(d, e, a, b, c, 77);
    R4(c, d, e, a, b, 78);
    R4(b, c, d, e, a, 79);
    /* Add the working vars back into context.state[] */
    state[0] += a;
    state[1] += b;
    state[2] += c;
    state[3] += d;
    state[4] += e;
    /* Wipe variables */
    a = b = c = d = e = 0;
#ifdef SHA1HANDSOFF
    memset(block, '\0', sizeof(block));
#endif
}


/* SHA1Init - Initialize new context */

void SHA1Init(
    SHA1_CTX * context
)
{
    /* SHA1 initialization constants */
    context->state[0] = 0x67452301;
    context->state[1] = 0xEFCDAB89;
    context->state[2] = 0x98BADCFE;
    context->state[3] = 0x10325476;
    context->state[4] = 0xC3D2E1F0;
    context->count[0] = context->count[1] = 0;
}


/* Run your data through this. */

void SHA1Update(
    SHA1_CTX * context,
    const unsigned char *data,
    uint32_t len
)
{
    uint32_t i;

    uint32_t j;

    j = context->count[0];
    if ((context->count[0] += len << 3) < j)
        context->count[1]++;
    context->count[1] += (len >> 29);
    j = (j >> 3) & 63;
    if ((j + len) > 63)
    {
        memcpy(&context->buffer[j], data, (i = 64 - j));
        SHA1Transform(context->state, context->buffer);
        for (; i + 63 < len; i += 64)
        {
            SHA1Transform(context->state, &data[i]);
        }
        j = 0;
    }
    else
        i = 0;
    memcpy(&context->buffer[j], &data[i], len - i);
}


/* Add padding and return the message digest. */

void SHA1Final(
    unsigned char digest[20],
    SHA1_CTX * context
)
{
    unsigned i;

    unsigned char finalcount[8];

    unsigned char c;

#if 0    /* untested "improvement" by DHR */
    /* Convert context->count to a sequence of bytes
     * in finalcount.  Second element first, but
     * big-endian order within element.
     * But we do it all backwards.
     */
    unsigned char *fcp = &finalcount[8];

    for (i = 0; i < 2; i++)
    {
        uint32_t t = context->count[i];

        int j;

        for (j = 0; j < 4; t >>= 8, j++)
            *--fcp = (unsigned char) t}
#else
    for (i = 0; i < 8; i++)
    {
        finalcount[i] = (unsigned char) ((context->count[(i >= 4 ? 0 : 1)] >> ((3 - (i & 3)) * 8)) & 255);      /* Endian independent */
    }
#endif
    c = 0200;
    SHA1Update(context, &c, 1);
    while ((context->count[0] & 504) != 448)
    {
        c = 0000;
        SHA1Update(context, &c, 1);
    }
    SHA1Update(context, finalcount, 8); /* Should cause a SHA1Transform() */
    for (i = 0; i < 20; i++)
    {
        digest[i] = (unsigned char)
            ((context->state[i >> 2] >> ((3 - (i & 3)) * 8)) & 255);
    }
    /* Wipe variables */
    memset(context, '\0', sizeof(*context));
    memset(&finalcount, '\0', sizeof(finalcount));
}

void SHA1(
    char *hash_out,
    const char *str,
    int len)
{
    SHA1_CTX ctx;
    unsigned int ii;

    SHA1Init(&ctx);
    for (ii=0; ii<len; ii+=1)
        SHA1Update(&ctx, (const unsigned char*)str + ii, 1);
    SHA1Final((unsigned char *)hash_out, &ctx);
    hash_out[20] = '\0';
}




static char encoding_table[] = {'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',
                                'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',
                                'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X',
                                'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f',
                                'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n',
                                'o', 'p', 'q', 'r', 's', 't', 'u', 'v',
                                'w', 'x', 'y', 'z', '0', '1', '2', '3',
                                '4', '5', '6', '7', '8', '9', '+', '/'};
static char *decoding_table = NULL;
static int mod_table[] = {0, 2, 1};

void build_decoding_table() {
     int i;
    decoding_table = malloc(256);
 
    for (i = 0; i < 64; i++)
        decoding_table[(unsigned char) encoding_table[i]] = i;
}
 
 
void base64_cleanup() {
    free(decoding_table);
} 
 
char *base64_encode(const unsigned char *data,
                    size_t input_length,
                    size_t *output_length) {
    int i, j;
    char* encoded_data;
    *output_length = 4 * ((input_length + 2) / 3);
 
    encoded_data = malloc(*output_length);
    if (encoded_data == NULL) return NULL;
 
    for (i = 0, j = 0; i < input_length;) {
 
        uint32_t octet_a = i < input_length ? (unsigned char)data[i++] : 0;
        uint32_t octet_b = i < input_length ? (unsigned char)data[i++] : 0;
        uint32_t octet_c = i < input_length ? (unsigned char)data[i++] : 0;
 
        uint32_t triple = (octet_a << 0x10) + (octet_b << 0x08) + octet_c;
 
        encoded_data[j++] = encoding_table[(triple >> 3 * 6) & 0x3F];
        encoded_data[j++] = encoding_table[(triple >> 2 * 6) & 0x3F];
        encoded_data[j++] = encoding_table[(triple >> 1 * 6) & 0x3F];
        encoded_data[j++] = encoding_table[(triple >> 0 * 6) & 0x3F];
    }
 
    for (i = 0; i < mod_table[input_length % 3]; i++)
        encoded_data[*output_length - 1 - i] = '=';
 
    return encoded_data;
}
 
 
unsigned char *base64_decode(const char *data,
                             size_t input_length,
                             size_t *output_length) {
    int i, j;
    unsigned char *decoded_data;
    if (decoding_table == NULL) build_decoding_table();
 
    if (input_length % 4 != 0) return NULL;
 
    *output_length = input_length / 4 * 3;
    if (data[input_length - 1] == '=') (*output_length)--;
    if (data[input_length - 2] == '=') (*output_length)--;
 
    decoded_data = malloc(*output_length);
    if (decoded_data == NULL) return NULL;
 
    for (i = 0, j = 0; i < input_length;) {
 
        uint32_t sextet_a = data[i] == '=' ? 0 & i++ : decoding_table[data[i++]];
        uint32_t sextet_b = data[i] == '=' ? 0 & i++ : decoding_table[data[i++]];
        uint32_t sextet_c = data[i] == '=' ? 0 & i++ : decoding_table[data[i++]];
        uint32_t sextet_d = data[i] == '=' ? 0 & i++ : decoding_table[data[i++]];
 
        uint32_t triple = (sextet_a << 3 * 6)
        + (sextet_b << 2 * 6)
        + (sextet_c << 1 * 6)
        + (sextet_d << 0 * 6);
 
        if (j < *output_length) decoded_data[j++] = (triple >> 2 * 8) & 0xFF;
        if (j < *output_length) decoded_data[j++] = (triple >> 1 * 8) & 0xFF;
        if (j < *output_length) decoded_data[j++] = (triple >> 0 * 8) & 0xFF;
    }
 
    return decoded_data;
}

int PortNumber;
int FirstParamsUpdate=1;
pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
int CurrentSection = 1;

   int a2 = 6300446;
   int b2 = 7501193;
   int c2 = 4316401;

   int next2()
   { 
       pthread_mutex_lock(&mutex);
       a2=a2*b2+c2; 
       if(a2<0) a2=-a2;
       CurrentSection++; 
       pthread_mutex_unlock(&mutex);
       return a2; 
   }


typedef struct wordlistelement
{
   char* word;
   struct wordlistelement *next;
} WordListElement;

void getFractions(char *pc, char *dst)
{
   char fraction[10];
   char *pfraction;
   *dst = '\0';
   while(1)
   {
      pfraction = strchr(pc, '/');
      if(pfraction==NULL)
         return;
      if(*dst!='\0')
         strcat(dst, " ");
      pfraction--;
      while(*pfraction>='0'&&*pfraction<='9')
         pfraction--;
      pfraction++;
      sscanf(pfraction, "%s", fraction);
      pc=pfraction+strlen(fraction);
      strcat(dst, fraction);
   }
}

WordListElement* splitL(char* phrase, int wordcnt, int *wordCntFound)
{
   WordListElement 
      *WordListHead = NULL, 
      *WordListTail = NULL, 
      *NewWord = NULL;
   int len = 0;
   char *blanks = " \t\n\r";
   char *pc = phrase;
   char *wordread = (char*)malloc(strlen(phrase)+1);
   if(wordread==NULL)
   {
      printf("E001: Memory allocation failed\n");
      return NULL;
   }
   int proceed = 1;
   while (proceed&&*pc!='\0')
   {
      while(*pc!='\0'&&strchr(blanks, *pc)!=NULL)
         pc++;
      if(*pc=='\0')
         break;
      if(len<wordcnt)
         sscanf(pc, "%s", wordread);
      else
      {
         getFractions(pc, wordread);
         proceed = 0;
      }
      pc+=strlen(wordread);
      NewWord = (WordListElement*)malloc(sizeof(WordListElement));
      if(NewWord==NULL)
      {
         printf("E002: Memory allocation failed\n");
         return NULL;
      }
      NewWord->word = (char*)malloc(sizeof(strlen(wordread)+1));
      if(NewWord->word==NULL)
      {
         free(NewWord);
         printf("E003: Memory allocation failed\n");
         return NULL;
      }
      strcpy(NewWord->word, wordread);
      NewWord->next = NULL;
      if(WordListTail!=NULL)
         WordListTail->next = NewWord;
      WordListTail = NewWord;
      if(WordListHead == NULL)
         WordListHead = NewWord;
      len++;
   }
   free(wordread);
   *wordCntFound = len;
   return WordListHead;
}

WordListElement* split(char* phrase, int *wordcnt)
{
   WordListElement 
      *WordListHead = NULL, 
      *WordListTail = NULL, 
      *NewWord = NULL;
   int len = 0;
   char *blanks = " \t\n\r";
   char *pc = phrase;
   char *wordread = (char*)malloc(strlen(phrase)+1);
   if(wordread==NULL)
   {
      printf("E004: Memory allocation failed\n");
      return NULL;
   }
   while (*pc!='\0')
   {
      while(*pc!='\0'&&strchr(blanks, *pc)!=NULL)
         pc++;
      if(*pc=='\0')
         break;
      sscanf(pc, "%s", wordread);
      pc+=strlen(wordread);
      NewWord = (WordListElement*)malloc(sizeof(WordListElement));
      if(NewWord==NULL)
      {
         printf("E005: Memory allocation failed\n");
         return NULL;
      }
      NewWord->word = (char*)malloc(sizeof(strlen(wordread)+1));
      if(NewWord->word==NULL)
      {
         free(NewWord);
         printf("E006: Memory allocation failed\n");
         return NULL;
      }
      strcpy(NewWord->word, wordread);
      NewWord->next = NULL;
      if(WordListTail!=NULL)
         WordListTail->next = NewWord;
      WordListTail = NewWord;
      if(WordListHead == NULL)
         WordListHead = NewWord;
      len++;
   }
   *wordcnt = len;
   return WordListHead;
}

void WordListCleanup(WordListElement *head, int deep)
{
   WordListElement *kill = head;
   while(kill!=NULL)
   {
      if(deep)
         free(head->word);
      head = head->next;
      free(kill);
      kill = head;
   }
}

WordListElement *wordlists[10];

void WordListArrayCleanup( int len)
{
   int i;
   for(i=0;i<len;i++)
      WordListCleanup(wordlists[i], 1);
}

typedef struct clientlistelement
{
   int index;
   char *Name;
   int NewState;
   int Socket;
   char *Melody;
   char sessionkey[50];
   int protocol;
   char *handshakeresponse;
   int playing;
   struct clientlistelement *next;
} ClientListElement;

ClientListElement *ClientList = NULL, *ClientListTail = NULL;
ClientListElement *MonitorList = NULL, *MonitorListTail = NULL;
int clientCount = 0;
int monitorCount = 0;

char *Params = NULL;
char *Melody = NULL;

int listening = 1;
int l;

ClientListElement * findInsertionPlace()
{
   int index = 0;
   ClientListElement* currClient=ClientList;
   if(currClient==NULL||currClient->index>0)
      return NULL;
   while(currClient!=NULL)
   {
      if(currClient->next!=NULL&&currClient->next->index>index+1)
         return currClient;
      currClient=currClient->next;
      index++;
   }
   return currClient;
}

ClientListElement* createClient(int socket)
{
   ClientListElement* newClient;
   newClient = (ClientListElement*)malloc(sizeof(ClientListElement));
   if(newClient==NULL)
   {
      printf("E007: Memory allocation failed\n");
      return NULL;
   }
   newClient->Socket = socket;
   newClient->next = NULL;
   newClient->Name = (char*)malloc(8);
   if(newClient->Name==NULL)
   {
      printf("E008: Memory allocation failed\n");
      free(newClient);
      return NULL;
   }
   strcpy(newClient->Name, "noname");
   newClient->Melody = (char*)malloc(8);
   if(newClient->Melody==NULL)
   {
      printf("E009: Memory allocation failed\n");
      free(newClient->Name);
      free(newClient);
      return NULL;
   }
   strcpy(newClient->Melody, "0 0 0 0");
   newClient->NewState = 1;
   newClient->playing = 1;
   newClient->handshakeresponse = NULL;
   return newClient;
}

ClientListElement *addMonitor(ClientListElement *newMonitor)
{
   pthread_mutex_lock(&mutex);
   if(MonitorListTail==NULL)
   {
      newMonitor->index = 0;
      MonitorList = newMonitor;
   }
   else
   {
      newMonitor->index = MonitorListTail->index+1;
      MonitorListTail->next = newMonitor;
   }
   MonitorListTail = newMonitor;
   monitorCount++;
   printf("There are %d Monitors.\n", monitorCount);      
   pthread_mutex_unlock(&mutex);
   return newMonitor;
}

ClientListElement* addClient(ClientListElement *newClient)
{
   ClientListElement *InsertionPlace;
   pthread_mutex_lock(&mutex);
   InsertionPlace = findInsertionPlace();
   if(InsertionPlace!=NULL)
      newClient->index = InsertionPlace->index+1;
   else
   {
      if(ClientListTail==NULL)
         newClient->index = 0;
      else
      if(ClientList->index>0)
         newClient->index = 0;
      else
         newClient->index = ClientListTail->index+1;
   }
   if(InsertionPlace == NULL)
   {
      if(ClientListTail!=NULL)
      {
         if(ClientList->index>0)
         {
            newClient->next = ClientList;
            ClientList = newClient;
         }
         else
         {
            ClientListTail->next = newClient;
            ClientListTail = newClient;
         }
      }
      else
      {
         ClientList = newClient;
         ClientListTail = newClient;
      }
   }
   else
   {
      newClient->next = InsertionPlace->next;
      InsertionPlace->next = newClient;
   }
   clientCount++;
   printf("There are %d clients.\n", clientCount);      
   pthread_mutex_unlock(&mutex);
   return newClient;
}

void removeClient(int i)
{
   pthread_mutex_lock(&mutex);
   ClientListElement *currClient = ClientList;
   ClientListElement *prev = ClientList;
   while(currClient!=NULL)
   {
      if(currClient->index == i)
      {
         if(prev==currClient)
            ClientList = ClientList->next;
         else
            prev->next = currClient->next;
         if(currClient->Name!=NULL)
            free(currClient->Name);
         if(currClient->Melody!=NULL)
            free(currClient->Melody);
         if(currClient->handshakeresponse!=NULL)
            free(currClient->handshakeresponse);
         free(currClient);
         clientCount--;
         printf("There are %d players left.\n", clientCount);
         break;
      }
      prev = currClient;
      currClient = currClient->next;
   }
   if(ClientList==NULL)
      ClientListTail = NULL;
   else
   if(prev->next==NULL)
      ClientListTail = prev;
   pthread_mutex_unlock(&mutex);
}

void removeMonitor(int i)
{
   pthread_mutex_lock(&mutex);
   ClientListElement *currMonitor = MonitorList;
   ClientListElement *prev = MonitorList;
   while(currMonitor!=NULL)
   {
      if(currMonitor->index == i)
      {
         if(prev==currMonitor)
            MonitorList = MonitorList->next;
         else
            prev->next = currMonitor->next;
         if(currMonitor->Name!=NULL)
            free(currMonitor->Name);
         if(currMonitor->Melody!=NULL)
            free(currMonitor->Melody);
         if(currMonitor->handshakeresponse!=NULL)
            free(currMonitor->handshakeresponse);
         free(currMonitor);
         monitorCount--;
         printf("There are %d monitors left.\n", monitorCount);
         break;
      }
      prev = currMonitor;
      currMonitor = currMonitor->next;
   }
   if(MonitorList==NULL)
      MonitorListTail = NULL;
   else
   if(prev->next==NULL)
      MonitorListTail = prev;
   pthread_mutex_unlock(&mutex);
}

void SendMessageToPlayer(char* message, ClientListElement* player)
{
   char *response = (char*)malloc(strlen(message)+3);
   response[0] = (char)(-127);
   response[1] = (char)(strlen(message));
   memcpy(response+2, message, strlen(message)+1);
   write(player->Socket , response , strlen(message)+2);
   free (response);
}

void SendAllNamesToMonitor(ClientListElement *monitor)
{
   ClientListElement* currClient = ClientList;
   char message[120];
   char *msgptr = message+1;
   sprintf(message, "=");
   while(currClient!=NULL)
   {
      sprintf(msgptr, " %d %s", currClient->index, currClient->Name);
      msgptr=strchr(msgptr, '\0');
      currClient = currClient->next;   
   }
   SendMessageToPlayer(message, monitor);
}

void BroadcastNewNameMonitors(int index, char *name)
{
   char *response;
   ClientListElement* currClient;
   char* message = (char*)malloc(strlen(name)+5);
   sprintf(message, "# %d %s", index, name);
   response = (char*)malloc(strlen(message)+3);
   response[0] = (char)(-127);
   response[1] = (char)(strlen(message));
   memcpy(response+2, message, strlen(message)+1);
   currClient = MonitorList;
   while(currClient!=NULL)
   {
      write(currClient->Socket , response , strlen(message)+2);
      currClient = currClient->next;
   }
   free (response);
   free (message);
}

void BroadcastMessage(char *message)
{
   char *response = (char*)malloc(strlen(message)+3);
   ClientListElement* currClient;
   response[0] = (char)(-127);
   response[1] = (char)(strlen(message));
   memcpy(response+2, message, strlen(message)+1);
   currClient = ClientList;
   while(currClient!=NULL)
   {
      write(currClient->Socket , response , strlen(message)+2);
      currClient = currClient->next;
   }
   currClient = MonitorList;
   while(currClient!=NULL)
   {
      write(currClient->Socket , response , strlen(message)+2);
      currClient = currClient->next;
   }
   free (response);
}

void BroadcastSection()
{
   char *response = (char*)malloc(50);
   ClientListElement* currClient;
   response[0] = (char)(-127);
   sprintf(response+2, "+ %d %d", a2, CurrentSection);
   response[1] = (char)(strlen(response+2));
   currClient = ClientList;
   while(currClient!=NULL)
   {
      write(currClient->Socket , response , strlen(response+2)+2);
      currClient = currClient->next;
   }
   currClient = MonitorList;
   while(currClient!=NULL)
   {
      write(currClient->Socket , response , strlen(response+2)+2);
      currClient = currClient->next;
   }
   free (response);
}

int UpdateMelodyAndBroadcast()
{
   int phrasecount;
   int i, maxlen=0, stop = 0, blockcount = 1, blocklen = 1024;
   int* lengths;
   WordListElement *currentword[10];
   char *CombinedPhrase = NULL;
   ClientListElement* currClient = ClientList;
   int beginning = 1;
   pthread_mutex_lock(&mutex);
   phrasecount = clientCount;
   
   lengths = (int*)malloc(phrasecount*sizeof(int));
   if(lengths==NULL)
   {
      printf("E013: Memory allocation failed\n");
      pthread_mutex_unlock(&mutex);
      return 1;
   }
   CombinedPhrase = (char*)malloc(blocklen);
   if(CombinedPhrase==NULL)
   {
      printf("E014: Memory allocation failed\n");
      free(lengths);
      pthread_mutex_unlock(&mutex);
      return 1;
   }
   CombinedPhrase[0] = '\0';

   for(i=0;i<phrasecount;i++)
   {
      wordlists[i]=split(currClient->Melody, &lengths[i]);
      currentword[i] = wordlists[i];
      if(lengths[i]>maxlen)
         maxlen = lengths[i];
      currentword[i] = wordlists[i];
      currClient = currClient->next;
   }
   do
   {
      for(i=0;i<phrasecount;i++)
      {
//+3 is for ending zero, blank and eventual new line if this is
//the last addition
         if(strlen(CombinedPhrase)+3+strlen(currentword[i]->word)
            >=blocklen*blockcount)
         {
            char* oldCombinedPhrase = CombinedPhrase;
            blockcount++;
            CombinedPhrase = (char*)malloc(blocklen*blockcount);
            if(CombinedPhrase==NULL)
            {
               printf("E015: Memory allocation failed\n");
               free(lengths);
               free(oldCombinedPhrase);
               pthread_mutex_unlock(&mutex);
               return 1;
            }
            strcpy(CombinedPhrase, oldCombinedPhrase);
            free(oldCombinedPhrase);
         }
         if(!beginning)
            strcat(CombinedPhrase, " ");
         else
            beginning = 0;
         strcat(CombinedPhrase, currentword[i]->word);
         currentword[i] = currentword[i]->next;
         if(currentword[i]==NULL)
            currentword[i] = wordlists[i];
      }
      stop++;
   }
   while(stop<maxlen);
   WordListArrayCleanup(phrasecount);
   if(Melody != NULL)
      free(Melody);
   Melody = CombinedPhrase;  
   BroadcastMessage(Melody);
   pthread_mutex_unlock(&mutex);
   return 0;
}

#define PROTOCOL_PLAYER 0
#define PROTOCOL_MONITOR 1

char *Protocols[] = {"Player", "Monitor"};

#define MAXMESSAGE 5000

int getProtocol(char *msg)
{
   char *pProtocol;
   char *pProtocolLineStart = strstr(msg, "Sec-WebSocket-Protocol: ");
   if(pProtocolLineStart==NULL)
      return PROTOCOL_PLAYER;
   pProtocol = pProtocolLineStart+strlen("Sec-WebSocket-Protocol: ");
   if(strncmp( pProtocol, "Monitor", 7)==0 )
      return PROTOCOL_MONITOR;
   return PROTOCOL_PLAYER;
}

int handshake(char *msg, ClientListElement* myClient)
{
   int i;
   char keywithdigest[100];
   char digest[50];
   char *pkey;
   char *pkeylinestart = NULL;
   char *base64encodeddigest = NULL;
   size_t encodeddigestsize;
   pkeylinestart = strstr(msg, "Sec-WebSocket-Key: ");
   if(pkeylinestart==NULL)
   {
      printf("Error in handshake");
      return 0;
   }
   pkey = pkeylinestart + strlen("Sec-WebSocket-Key: ");
   sscanf(pkey, "%s", myClient->sessionkey);
   sprintf(keywithdigest, "%s258EAFA5-E914-47DA-95CA-C5AB0DC85B11", myClient->sessionkey);
   
   SHA1(digest, (const unsigned char *) keywithdigest, strlen(keywithdigest));
   myClient->handshakeresponse = base64_encode((const unsigned char*)digest, 20, &encodeddigestsize);
   myClient->protocol = getProtocol(msg);
   if(myClient->protocol == PROTOCOL_PLAYER)
      addClient(myClient);
   else
      addMonitor(myClient);
   return 1;
}

void handshake_response(ClientListElement* myClient)
{
   int sock = myClient->Socket;
   
    char firstline[]="HTTP/1.1 101 Switching Protocols\r\n";
    char secondline[]="Upgrade: websocket\r\n";
    char thirdline[]="Connection: Upgrade\r\n";
    char fourthline[100]="Sec-WebSocket-Accept: ";
    char fifthline[100]="Sec-WebSocket-Protocol: ";
    char sixthline[]="\r\n";
    
    strcat(fourthline, myClient->handshakeresponse);
    strcat(fourthline, "\r\n");
    strcat(fifthline, Protocols[myClient->protocol]);
    strcat(fifthline, "\r\n");
    
    write(sock, firstline, strlen(firstline));
    write(sock, secondline, strlen(secondline));
    write(sock, thirdline, strlen(thirdline));
    write(sock, fourthline, strlen(fourthline));
    write(sock, fifthline, strlen(fifthline));
    write(sock, sixthline, strlen(sixthline));
    
}

char *trim(char *msg)
{
   char *start = msg;
   char *end = msg+strlen(msg)-1;
   char *result = NULL;
   while(start<msg+strlen(msg)&&(*start==' '||*start=='\r'||*start=='\t'||*start=='\n'))
      start++;
   while(end>msg&&(*end==' '||*end=='\r'||*end=='\t'||*end=='\n'))
      end--;
   result = (char*)malloc(end-start+2);
   memcpy(result, start, end-start+1);
   result[end-start+1]='\0';
   free(msg);
   return result;
}

/*
 * This will handle connection for each client
 * */
void *connection_handler(void *client)
{
   int i, j;
   ClientListElement* myClient = (ClientListElement*)client;
   //Get the socket descriptor
   int sock = myClient->Socket;
   int read_size;

   char client_message[MAXMESSAGE];
   client_message[0]='\0';

   read_size = recv(sock , client_message , MAXMESSAGE, 0);
   if(read_size>0)
   {
      char *welcome_message = NULL;
      if(!handshake(client_message, myClient))
      {
         printf("E024: Websocket handshake failed\n");
         return 0;
      }
      handshake_response(myClient);

      if(myClient->Name!=NULL)
         free(myClient->Name);
      myClient->Name = (char*)malloc(10);
      if(myClient->Name==NULL)
      {
         printf("E019: Memory allocation failed\n");
         return 0;
      }
      sprintf(myClient->Name, "%s%d", Protocols[myClient->protocol], myClient->index);
      printf("New %s: %s\n", 
         Protocols[myClient->protocol], myClient->Name);
      UpdateMelodyAndBroadcast();
      while(myClient->playing)
      {
         int msglen;
         char *decoded;
         read_size = recv(sock , client_message , MAXMESSAGE, 0);
         if(read_size<=0)
            break;
         *(client_message+read_size)='\0';
         msglen = (int)(client_message[1]&0xFF)-128;
         if(msglen>0)
         {
            decoded = (char*)malloc(msglen+1);
            for(j=0;j<msglen;j++)
               decoded[j]=(char)(client_message[j+6]^client_message[(j & 0x3)+2]);
            decoded[j]='\0';
            decoded = trim(decoded);
            if(myClient->protocol==PROTOCOL_MONITOR)
            {
               if(msglen==1&&decoded[0]=='?'&&ClientList!=NULL)
               {
                  SendAllNamesToMonitor(myClient);
                  continue;
               }
               else
               if(msglen==2&&decoded[0]==3&&(decoded[1]&0xFF)==233)
               {
                  myClient->playing = 0;
                  break;
               }
               else 
                  continue;
            }
            //From here on it's a Player
            if(decoded[0]=='\0')
            {
            }
            else
            if(msglen==2&&decoded[0]==3&&(decoded[1]&0xFF)==233)
            {
               myClient->playing = 0;
               break;
            }
            else
            if(msglen==1&&decoded[0]=='+')//new section
            {
               next2();
               BroadcastSection();
            }
            else
            if(msglen>2&&decoded[0]=='?')//get index
            {
               int NameLen = strlen(decoded+2);
               char *name = (char*)malloc(NameLen+1);
               char myIndex[10];
               strcpy(name, decoded+2);
               if(myClient->Name!=NULL)
                  free(myClient->Name);
               myClient->Name = name;
               sprintf(myIndex, "%d", myClient->index);
               SendMessageToPlayer(myIndex, myClient);
               BroadcastNewNameMonitors(myClient->index, myClient->Name);
            }
            else
            if(decoded[0]<'0'||decoded[0]>'9')
            {
               if(Params!=NULL)
                  free(Params);
               Params = (char*)malloc(strlen(decoded)+1);
               strcpy(Params, decoded);
               BroadcastMessage(Params);
            }
            else
            {
               if(myClient->Melody!=NULL)
                  free(myClient->Melody);
               myClient->Melody = (char*)malloc(strlen(decoded)+1);
               strcpy(myClient->Melody, decoded);
               UpdateMelodyAndBroadcast();
            }
         }
         else
            printf("E26: Error in data length.\n");
      }
   }
   printf("%s disconnected\n", myClient->Name);
   if(myClient->protocol==PROTOCOL_PLAYER)
      removeClient(myClient->index);
   else
      removeMonitor(myClient->index);
   UpdateMelodyAndBroadcast();
   fflush(stdout);
   if(ClientList==NULL)
      printf("no players\n");
   return 0;
}

int LoadPortNumber()
{
   int retval;
   FILE *f = fopen("ServerConfiguration.txt", "rt");
   if(f==NULL)
   {
      printf("Could not open file ServerConfiguration.txt\n");
      return -1;
   }
   retval = fscanf(f, "%d", &PortNumber);
   if(retval!=1||PortNumber<=0)
   {
      printf("Invalid file ServerConfiguration.txt\n");
      fclose(f);
      return -1;
   }
   fclose(f);
   printf("Port number: %d\n", PortNumber);
   return 0;
}

/*
 * This will handle connection for each client
 * */
void *starter_handler(void *dummy)
{
   int socket_desc , new_socket , c;
   struct sockaddr_in server , client;
   if(LoadPortNumber()!=0)
      return 0;
   printf("Server started.\n");
   //Create socket
   socket_desc = socket(AF_INET , SOCK_STREAM , 0);
   if (socket_desc == -1)
   {
      printf("Could not create socket");
      return 0;
   }
   
   //Prepare the sockaddr_in structure
   server.sin_family = AF_INET;
   server.sin_addr.s_addr = INADDR_ANY;
   server.sin_port = htons( PortNumber );
   
   //Bind
   if( bind(socket_desc,(struct sockaddr *)&server , sizeof(server))  < 0)
   {
      puts("bind failed");
      return 0;
   }
   
   //Listen
   listen(socket_desc , 3);
   

   //Accept and incoming connection
   puts("Waiting for incoming connections...");
   c = sizeof(struct sockaddr_in);
   while( (new_socket = accept(socket_desc, (struct sockaddr *)&client, (socklen_t*)&c)) )
   {

      ClientListElement *newClient = NULL;
      if(listening==0)
         break;

      if(clientCount>=9)
      {
         write(new_socket, "Too many clients. 9 is the maximum.\n", 36);
         continue;
      } 
      
      pthread_t client_thread;  
      
      newClient = createClient(new_socket);

      if(newClient==NULL)
         continue;

      if( pthread_create( &client_thread , NULL ,  connection_handler , (void*) newClient)  < 0)
      {
         perror("could not create thread");
         return 0;
      }
      
   }
   
   if (new_socket <0)
   {
      perror("accept failed");
      return 0;
   }
   
   return 0;
}

int main(int argc , char *argv[])
{
   pthread_t starter_thread;

   a2 = time(NULL);

   if( pthread_create( &starter_thread , NULL ,  starter_handler , NULL)  < 0)
   {
      perror("could not create stopper thread");
      return 1;
   }

   int inChar = '\0';
   printf("Press '0' to stop the server (only if no clients are connected).\n");
   while(inChar!=48)
   {
      scanf("%c", &inChar);
      if(inChar!=48)
         continue;
      if(clientCount<=0)
         break;
      else
      {
         printf("There are still %d clients connected.\n", clientCount);
         inChar = '\0';
      }
   }
}

